#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""
Service functions for logging

Configure console and file logging; Create and handle config file for api keys;

SPDX-License-Identifier: AGPL-3.0-or-later
"""

__copyright__ = "© Reiner Lemoine Institut"
__license__ = "GNU Affero General Public License Version 3 (AGPL-3.0)"
__url__ = "https://www.gnu.org/licenses/agpl-3.0.en.html"
__author__ = "Ludee; christian-rli"
__issue__ = "https://github.com/OpenEnergyPlatform/examples/issues/52"
__version__ = "v0.10.0"

import os
import yaml
import shutil
import pathlib
from datetime import date

import logging
import logging.config
from open_mastr.utils.constants import (
    TECHNOLOGIES,
    API_LOCATION_TYPES,
    ADDITIONAL_TABLES,
)


log = logging.getLogger(__name__)


def get_project_home_dir():
    """Get root dir of project data

    On linux this path equals `$HOME/.open-MaStR/`, respectively `~/.open-MaStR/`
    which is also called `PROJECTHOME`.

    Returns
    -------
    path-like object
        Absolute path to root dir of open-MaStR project home
    """

    return os.path.join(os.path.expanduser("~"), ".open-MaStR")


def get_output_dir():
    """Get output directory for csv data, xml file and database. Defaults to get_project_home_dir()

    Returns
    -------
    path-like object
        Absolute path to output path
    """

    if "OUTPUT_PATH" in os.environ:
        return os.environ.get("OUTPUT_PATH")

    return get_project_home_dir()


def get_data_version_dir():
    """
    Subdirectory of data/ in PROJECTHOME

    See :ref:`docs <Project directory>` for configuration of data version.

    Returns
    -------
    path-like object
        Absolute path to `PROJECTHOME/data/<data-version>/`
    """
    data_version = get_data_config()

    if "OUTPUT_PATH" in os.environ:
        return os.path.join(os.environ.get("OUTPUT_PATH"), "data", data_version)

    return os.path.join(get_project_home_dir(), "data", data_version)


def get_filenames():
    """
    Get file names defined in config

    Returns
    -------
    dict
        File names used in open-MaStR
    """
    with open(
        os.path.join(get_project_home_dir(), "config", "filenames.yml")
    ) as filename_fh:
        filenames = yaml.safe_load(filename_fh)

    return filenames


def get_data_config():
    """
    Get data version

    Returns
    -------
    str
        dataversion
    """

    today = date.today()

    data_config = f'dataversion-{today.strftime("%Y-%m-%d")}'

    return data_config


def create_project_home_dir():
    """Create directory structure of PROJECTHOME"""
    project_home = get_project_home_dir()

    # Create root project home path
    if not os.path.isdir(project_home):
        # Create project home
        log.info(f"Create {project_home} used for config, parameters and data.")
        os.mkdir(project_home)

    # Create project home subdirs
    subdirs = [os.path.join(project_home, s) for s in ["config", "logs", "data"]]
    for subdir in subdirs:
        if not os.path.isdir(subdir):
            os.mkdir(subdir)

    # copy default config files
    config_path = os.path.join(get_project_home_dir(), "config")
    log.info(f"I will create a default set of config files in {config_path}")

    internal_config_dir = os.path.join(
        pathlib.Path(__file__).parent.absolute(), "config"
    )
    files = ["logging.yml"]

    for file in files:
        if file not in os.listdir(config_path):
            shutil.copy(
                os.path.join(internal_config_dir, file),
                os.path.join(config_path, os.path.basename(file)),
            )


def create_data_dir():
    """
    Create direct for current data version

    The directory that is created for this fata version can
    be returned by :func:`~.get_data_version_dir`.
    """

    os.makedirs(get_data_version_dir(), exist_ok=True)


def _filenames_generator():
    """Write default file names .yml to project home dir"""

    filenames_file = os.path.join(get_project_home_dir(), "config", "filenames.yml")

    # How files are prefixed
    prefix = "bnetza_mastr"

    # Additional data available for certain technologies
    type_specific_data = {
        "eeg": ["wind", "hydro", "solar", "biomass", "gsgk", "storage"],
        "kwk": ["combustion", "biomass", "gsgk"],
        "permit": [
            "wind",
            "biomass",
            "combustion",
            "gsgk",
            "nuclear",
            "solar",
            "hydro",
        ],
    }

    # Template for file names
    filenames_template = {
        "raw": {
            "joined": "{prefix}_{technology}_raw.csv",
            "basic": "{prefix}_{technology}_basic.csv",  # power-unit
            "extended": "{prefix}_{technology}_extended.csv",  # unit
            "eeg": "{prefix}_{technology}_eeg.csv",
            "kwk": "{prefix}_{technology}_kwk.csv",
            "permit": "{prefix}_{technology}_permit.csv",
            "extended_fail": "{prefix}_{technology}_extended_fail.csv",
            "eeg_fail": "{prefix}_{technology}_eeg_fail.csv",
            "kwk_fail": "{prefix}_{technology}_kwk_fail.csv",
            "permit_fail": "{prefix}_{technology}_permit_fail.csv",
        }
    }

    filenames = {}

    # Define filenames .yml with a dict
    for section, section_filenames in filenames_template.items():
        filenames[section] = {}
        for tech in TECHNOLOGIES:

            # Files for all technologies
            files = ["joined", "basic", "extended", "extended_fail"]

            # Additional file for some technologies
            for addit_data_type, techs in type_specific_data.items():
                if tech in techs:
                    files.append(addit_data_type)
                    files.append(addit_data_type + "_fail")

                # Create filename dictionary for one technologies
                tmp = {
                    k: v.format(prefix=prefix, technology=tech)
                    for k, v in section_filenames.items()
                    if k in files
                }

                # Collect file names for all technologies
                filenames[section].update({tech: tmp})

    # Add file names of cleaned data
    filenames["cleaned"] = {
        tech: f"{prefix}_{tech}_cleaned.csv" for tech in TECHNOLOGIES
    }

    # Add file names of processed data
    filenames["postprocessed"] = {tech: f"{prefix}_{tech}.csv" for tech in TECHNOLOGIES}

    # Add filenames for location data
    filenames["raw"].update(
        {loc: f"{prefix}_{loc}_raw.csv" for loc in API_LOCATION_TYPES}
    )

    # Add filenames for additional tables
    filenames["raw"].update(
        {
            "additional_table": {
                addit_table: f"{prefix}_{addit_table}_raw.csv"
                for addit_table in ADDITIONAL_TABLES
            }
        }
    )

    # Add metadata file
    filenames["metadata"] = "datapackage.json"

    with open(filenames_file, "w") as outfile:
        yaml.dump(filenames, outfile)
    log.info(f"File names configuration saved to {filenames_file}")


def setup_project_home():
    """Create open-MaStR project home directory structure

    Create PROJECTHOME returned by :func:`~.get_project_home_dir`.
    In addition, default config files are copied to `PROJECTHOME/config/`.
    """

    # Create directory structure of project home dir
    create_project_home_dir()

    # Save default file names
    _filenames_generator()


def setup_logger():
    """Configure logging in console and log file.

    Returns
    -------
    logging.Logger
        Logger with two handlers: console and file.
    """

    # Read logging config
    with open(
        os.path.join(get_project_home_dir(), "config", "logging.yml")
    ) as filename_fh:
        logging_config = yaml.safe_load(filename_fh)

    # Add logfile location
    logging_config["handlers"]["file"]["filename"] = os.path.join(
        get_project_home_dir(), "logs", "open_mastr.log"
    )

    logging.config.dictConfig(logging_config)
    return logging.getLogger("open-MaStR")


def column_renaming():
    """
    Column renaming for CSV export of raw data

    Helps to export duplicate columns from different data sources.

    Returns
    -------
    dict
        Suffix and column to be suffixed keyed by data type.
    """
    return {
        "basic_data": {
            "columns": ["BestandsanlageMastrNummer"],
            "suffix": "basic",
        },
        "unit_data": {
            "columns": [
                "EinheitMastrNummer",
                "EegMastrNummer",
                "KwkMastrNummer",
                "GenMastrNummer",
                "SpeMastrNummer",
                "EinheitBetriebsstatus",
                "NichtVorhandenInMigriertenEinheiten",
                "Bruttoleistung",
            ],
            "suffix": "extended",
        },
        "eeg_data": {
            "columns": ["EegMastrNummer", "DatumLetzteAktualisierung", "Meldedatum"],
            "suffix": "eeg",
        },
        "kwk_data": {
            "columns": [
                "KwkMastrNummer",
                "Meldedatum",
                "Inbetriebnahmedatum",
                "DatumLetzteAktualisierung",
                "AnlageBetriebsstatus",
                "VerknuepfteEinheiten",
                "AusschreibungZuschlag",
            ],
            "suffix": "kwk",
        },
        "permit_data": {
            "columns": ["GenMastrNummer", "DatumLetzteAktualisierung", "Meldedatum"],
            "suffix": "permit",
        },
    }



import sys
import sqlite3
from os.path import expanduser
import os
import pandas as pd
import numpy as np
import pytest

from open_mastr.xml_download.utils_cleansing_bulk import (
    create_katalogwerte_from_bulk_download,
    replace_mastr_katalogeintraege,
)

# Check if xml file exists
_xml_file_exists = False
_xml_folder_path = os.path.join(expanduser("~"), ".open-MaStR", "data", "xml_download")
if os.path.isdir(_xml_folder_path):
    for entry in os.scandir(path=_xml_folder_path):
        if "Gesamtdatenexport" in entry.name:
            _xml_file_exists = True

_sqlite_folder_path = os.path.join(expanduser("~"), ".open-MaStR", "data", "sqlite")
_sqlite_file_path = os.path.join(_sqlite_folder_path, "open-mastr.db")
_sqlite_db_exists = bool(os.path.exists(_sqlite_file_path))

# Silence ValueError caused by logger https://github.com/pytest-dev/pytest/issues/5502
@pytest.fixture(autouse=True)
def capture_wrap():
    sys.stderr.close = lambda *args: None
    sys.stdout.close = lambda *args: None
    yield


@pytest.fixture(scope="module")
def con():
    con = sqlite3.connect(_sqlite_file_path)
    yield con
    con.close()


@pytest.fixture(scope="module")
def zipped_xml_file_path():
    zipped_xml_file_path = None
    for entry in os.scandir(path=_xml_folder_path):
        if "Gesamtdatenexport" in entry.name:
            zipped_xml_file_path = os.path.join(_xml_folder_path, entry.name)

    return zipped_xml_file_path


@pytest.mark.skipif(
    not _xml_file_exists, reason="The zipped xml file could not be found."
)
def test_replace_mastr_katalogeintraege(zipped_xml_file_path):
    df_raw = pd.DataFrame({"ID": [0, 1, 2], "Bundesland": [335, 335, 336]})
    df_replaced = pd.DataFrame(
        {"ID": [0, 1, 2], "Bundesland": ["Bayern", "Bayern", "Bremen"]}
    )
    pd.testing.assert_frame_equal(
        df_replaced, replace_mastr_katalogeintraege(zipped_xml_file_path, df_raw)
    )


@pytest.mark.skipif(
    not _xml_file_exists, reason="The zipped xml file could not be found."
)
def test_create_katalogwerte_from_bulk_download(zipped_xml_file_path):
    katalogwerte = create_katalogwerte_from_bulk_download(
        zipped_xml_file_path=zipped_xml_file_path
    )
    assert type(katalogwerte) == dict
    assert len(katalogwerte) > 1000
    assert type(list(katalogwerte.keys())[0]) == int



from open_mastr.mastr import Mastr
import os
import sqlalchemy
import pytest
from os.path import expanduser
import pandas as pd
from open_mastr.utils.constants import TRANSLATIONS

_xml_file_exists = False
_xml_folder_path = os.path.join(expanduser("~"), ".open-MaStR", "data", "xml_download")
if os.path.isdir(_xml_folder_path):
    for entry in os.scandir(path=_xml_folder_path):
        if "Gesamtdatenexport" in entry.name:
            _xml_file_exists = True


@pytest.fixture
def db():
    return Mastr()


@pytest.fixture
def db_path():
    return os.path.join(
        os.path.expanduser("~"), ".open-MaStR", "data", "sqlite", "mastr-test.db"
    )


@pytest.fixture
def db_translated(db_path):
    engine = sqlalchemy.create_engine(f"sqlite:///{db_path}")
    db_api = Mastr(engine=engine)

    db_api.download(date="existing", data=["wind", "hydro", "biomass", "combustion"])
    db_api.translate()

    return db_api


def test_Mastr_init(db):
    # test if folder structure exists
    assert os.path.exists(db.home_directory)
    assert os.path.exists(db._sqlite_folder_path)

    # test if engine and connection were created
    assert type(db.engine) == sqlalchemy.engine.Engine


@pytest.mark.skipif(
    not _xml_file_exists, reason="The zipped xml file could not be found."
)
def test_Mastr_translate(db_translated, db_path):
    # test if database was renamed correctly
    transl_path = db_path[:-3] + "-translated.db"
    assert os.path.exists(transl_path)

    # test if columns got translated
    inspector = sqlalchemy.inspect(db_translated.engine)
    table_names = inspector.get_table_names()

    for table in table_names:
        for column in inspector.get_columns(table):
            column = column["name"]
            assert column in TRANSLATIONS.values() or column not in TRANSLATIONS.keys()

    # test if new translated version replaces previous one
    db_translated.engine.dispose()
    engine = sqlalchemy.create_engine(f"sqlite:///{db_path}")
    db_empty = Mastr(engine=engine)
    db_empty.translate()

    for table in table_names:
        assert pd.read_sql(sql=table, con=db_empty.engine).shape[0] == 0



from bokeh.palettes import Category10_10 as palette

# import geoviews as gv
import bokeh

# gv.extension('bokeh')


def plotPowerPlants(df):
    # size marker according to gross power output
    iMaxSize = 30
    iMinSize = 10
    df["size"] = (df["Bruttoleistung"] - df["Bruttoleistung"].min()) / (
        df["Bruttoleistung"].max() - df["Bruttoleistung"].min()
    ) * (iMaxSize - iMinSize) + iMinSize

    # convert datetime to string
    df["date"] = df["Inbetriebnahmedatum"].dt.strftime("%Y-%m-%d")

    # define tooltips
    tt = [
        ("Name", "@Name"),
        ("Address", "@Standort, @Bundesland, @Land"),
        ("Commissioning date", "@date"),
        ("Type", "@Einheittyp"),
        ("Gross generation", "@Bruttoleistung MW"),
    ]
    hover_tool = bokeh.models.HoverTool(tooltips=tt)

    # choose background tiles
    overlay = gv.tile_sources.Wikipedia

    # markers will be colored according to Einheittyp
    groups = df["Einheittyp"].unique()
    colors = {
        "Wasser": palette[0],
        "Biomasse": palette[2],
        "Windeinheit": palette[4],
    }

    # create geoviews point objects
    for group in groups:
        df_group = df.loc[
            df["Einheittyp"] == group,
            [
                "Name",
                "Standort",
                "Bundesland",
                "Land",
                "date",
                "Einheittyp",
                "Bruttoleistung",
                "Laengengrad",
                "Breitengrad",
                "size",
            ],
        ]
        points = gv.Points(
            df_group, ["Laengengrad", "Breitengrad"], label=group
        ).options(
            aspect=2,
            responsive=True,
            tools=[hover_tool],
            size="size",
            active_tools=["wheel_zoom"],
            fill_alpha=0.6,
            fill_color=colors[group],
            line_color="white",
        )
        overlay = overlay * points

    # hide group when clicking on legend
    overlay.options(click_policy="hide", clone=False)

    # return figure
    return overlay



import os
from concurrent.futures import ProcessPoolExecutor, wait
from io import StringIO
from multiprocessing import cpu_count
from shutil import Error
from zipfile import ZipFile

import re
import lxml
import numpy as np
import pandas as pd
import sqlalchemy
from sqlalchemy import select, create_engine, inspect
from sqlalchemy.sql import text
from sqlalchemy.sql.sqltypes import Date, DateTime

from open_mastr.utils.config import setup_logger
from open_mastr.utils.helpers import data_to_include_tables
from open_mastr.utils.orm import tablename_mapping
from open_mastr.xml_download.utils_cleansing_bulk import cleanse_bulk_data


def write_mastr_xml_to_database(
    engine: sqlalchemy.engine.Engine,
    zipped_xml_file_path: str,
    data: list,
    bulk_cleansing: bool,
    bulk_download_date: str,
) -> None:
    """Write the Mastr in xml format into a database defined by the engine parameter."""
    print("Starting bulk download and data cleansing...")

    include_tables = data_to_include_tables(data, mapping="write_xml")
    threads_data = []

    with ZipFile(zipped_xml_file_path, "r") as f:
        files_list = correct_ordering_of_filelist(f.namelist())

        for file_name in files_list:
            xml_table_name = extract_xml_table_name(file_name)

            if not is_table_relevant(xml_table_name, include_tables):
                continue

            sql_table_name = extract_sql_table_name(xml_table_name)
            threads_data.append(
                (
                    file_name,
                    xml_table_name,
                    sql_table_name,
                    str(engine.url),
                    engine.url.password,
                    zipped_xml_file_path,
                    bulk_download_date,
                    bulk_cleansing,
                )
            )

    interleaved_files = interleave_files(threads_data)
    number_of_processes = get_number_of_processes()

    if number_of_processes > 0:
        with ProcessPoolExecutor(max_workers=number_of_processes) as executor:
            futures = [
                executor.submit(process_xml_file, *item) for item in interleaved_files
            ]
            for future in futures:
                future.result()
            wait(futures)
    else:
        for item in interleaved_files:
            process_xml_file(*item)

    print("Bulk download and data cleansing were successful.")


def get_number_of_processes():
    """Get the number of processes to use for the bulk download. Returns -1 if the user has not opted for the
    parallelized implementation. Otherwise, we recommend using the number of available CPUs - 1. If the user wants to
    use more processes, they can set the custom environment variable."""
    if "NUMBER_OF_PROCESSES" in os.environ:
        try:
            number_of_processes = int(os.environ.get("NUMBER_OF_PROCESSES"))
        except ValueError:
            print("Warning: Invalid value for NUMBER_OF_PROCESSES. Fallback to 1.")
            return 1
        if number_of_processes >= cpu_count():
            print(
                f"Warning: Your system supports {cpu_count()} CPUs. Using "
                f"more processes than available CPUs may cause excessive "
                f"context-switching overhead."
            )
        return number_of_processes
    if "USE_RECOMMENDED_NUMBER_OF_PROCESSES" in os.environ:
        return int(min(cpu_count() - 1, 4))
    return -1


def process_xml_file(
    file_name: str,
    xml_table_name: str,
    sql_table_name: str,
    connection_url: str,
    password: str,
    zipped_xml_file_path: str,
    bulk_download_date: str,
    bulk_cleansing: bool,
) -> None:
    """Process a single xml file and write it to the database."""
    try:
        # If set, the connection url obfuscates the password. We must replace the masked password with the actual password.
        if password:
            connection_url = re.sub(
                r"://([^:]+):\*+@", r"://\1:" + password + "@", connection_url
            )

        # Each process will create its own engine to ensure isolation and efficient resource management.
        # The connection url obfuscates the password. We must replace the masked password with the actual password.
        engine = create_efficient_engine(connection_url)
        with ZipFile(zipped_xml_file_path, "r") as f:
            print(f"Processing file '{file_name}'...")
            if is_first_file(file_name):
                print(f"Creating table '{sql_table_name}'...")
                create_database_table(engine, xml_table_name)
            df = read_xml_file(f, file_name)
            df = process_table_before_insertion(
                df,
                xml_table_name,
                zipped_xml_file_path,
                bulk_download_date,
                bulk_cleansing,
            )
            if engine.dialect.name == "sqlite":
                add_table_to_sqlite_database(df, xml_table_name, sql_table_name, engine)
            else:
                add_table_to_non_sqlite_database(
                    df, xml_table_name, sql_table_name, engine
                )
    except Exception as e:
        print(f"Error processing file '{file_name}': '{e}'")


def create_efficient_engine(connection_url: str) -> sqlalchemy.engine.Engine:
    """Create an efficient engine for the SQLite database."""
    is_sqlite = connection_url.startswith("sqlite://")

    connect_args = {}

    if is_sqlite:
        # Wait for max 5 minutes before timing out.
        connect_args["timeout"] = 300
        # Lock the database only it is necessary to improve concurrency and performance.
        connect_args["isolation_level"] = "DEFERRED"
        # Allow multiple threads to access the database.
        connect_args["check_same_thread"] = False
    else:
        # Wait for max 5 minutes before timing out.
        connect_args["connect_timeout"] = 300

    return create_engine(
        connection_url,
        connect_args=connect_args,
        # Before returning a connection from the pool, check if the connection is still valid.
        pool_pre_ping=True,
        # Max number of connections in the pool.
        pool_size=10,
        # Create up to 20 more connections when the demand for connections is high.
        max_overflow=20,
        # Recycle inactive connections after 180 seconds to prevent stale connections.
        pool_recycle=180,
        # Wait for 30 seconds before raising an exception when the pool is full.
        pool_timeout=30,
    )


def interleave_files(threads_data: list):
    """
    Multiple threads will process different files at once. If the files target the same table, the risk of a
    "database lock" error (i.e., 2 threads attempting to modify the same table at the same time) is increased.
    To reduce this probability, we can "interleave" the files based on the table they belong to.
    Example:
        Initial order: AnlagenEegSolar_1, AnlagenEegSolar_2, ..., AnlagenEegSpeicher_1, AnlagenEegSpeicher_2, ...
        Interleaved order: AnlagenEegSolar_1, AnlagenEegSpeicher_1, ..., AnlagenEegSolar_2, AnlagenEegSpeicher_2, ...
    """
    files_grouped_by_table = {}

    for item in threads_data:
        table_name = item[2]
        if table_name not in files_grouped_by_table:
            files_grouped_by_table[table_name] = []
        files_grouped_by_table[table_name].append(item)

    sorted_threads_data = []
    max_no_files_per_table = max(
        len(group) for group in files_grouped_by_table.values()
    )

    for idx in range(max_no_files_per_table):
        for table_name in files_grouped_by_table.keys():
            files = files_grouped_by_table[table_name]
            if idx < len(files):
                sorted_threads_data.append(files[idx])

    return sorted_threads_data


def extract_xml_table_name(file_name: str) -> str:
    """Extract the XML table name from the file name."""
    return file_name.split("_")[0].split(".")[0].lower()


def extract_sql_table_name(xml_table_name: str) -> str:
    """Extract the SQL table name from the xml table name."""
    return tablename_mapping[xml_table_name]["__name__"]


def is_table_relevant(xml_table_name: str, include_tables: list) -> bool:
    """Checks if the table contains relevant data and if the user wants to
    have it in the database."""
    # few tables are only needed for data cleansing of the xml files and contain no
    # information of relevance
    try:
        boolean_write_table_to_sql_database = (
            tablename_mapping[xml_table_name]["__class__"] is not None
        )
    except KeyError:
        print(
            f"Table '{xml_table_name}' is not supported by your open-mastr version and "
            f"will be skipped."
        )
        return False
    # check if the table should be written to sql database (depends on user input)
    include_count = include_tables.count(xml_table_name)

    return include_count == 1 and boolean_write_table_to_sql_database


def create_database_table(
    engine: sqlalchemy.engine.Engine, xml_table_name: str
) -> None:
    orm_class = tablename_mapping[xml_table_name]["__class__"]
    orm_class.__table__.drop(engine, checkfirst=True)
    orm_class.__table__.create(engine)


def is_first_file(file_name: str) -> bool:
    """check if the file name indicates that it is the first file from the table"""
    return (
        file_name.split(".")[0].split("_")[-1] == "1"
        or len(file_name.split(".")[0].split("_")) == 1
    )


def cast_date_columns_to_datetime(
    xml_table_name: str, df: pd.DataFrame
) -> pd.DataFrame:
    sqlalchemy_columnlist = tablename_mapping[xml_table_name][
        "__class__"
    ].__table__.columns.items()
    for column in sqlalchemy_columnlist:
        column_name = column[0]
        if is_date_column(column, df):
            # Convert column to datetime64, invalid string -> NaT
            df[column_name] = pd.to_datetime(df[column_name], errors="coerce")
    return df


def cast_date_columns_to_string(xml_table_name: str, df: pd.DataFrame) -> pd.DataFrame:
    column_list = tablename_mapping[xml_table_name][
        "__class__"
    ].__table__.columns.items()
    for column in column_list:
        column_name = column[0]

        if not (column[0] in df.columns and is_date_column(column, df)):
            continue

        df[column_name] = pd.to_datetime(df[column_name], errors="coerce")

        if type(column[1].type) is Date:
            df[column_name] = (
                df[column_name].dt.strftime("%Y-%m-%d").replace("NaT", None)
            )
        elif type(column[1].type) is DateTime:
            df[column_name] = (
                df[column_name].dt.strftime("%Y-%m-%d %H:%M:%S.%f").replace("NaT", None)
            )
    return df


def is_date_column(column, df: pd.DataFrame) -> bool:
    return type(column[1].type) in [Date, DateTime] and column[0] in df.columns


def correct_ordering_of_filelist(files_list: list) -> list:
    """Files that end with a single digit number get a 0 prefixed to this number
    to correct the list ordering. Afterwards the 0 is deleted again."""
    files_list_ordered = []
    count_if_zeros_are_prefixed = 0
    for file_name in files_list:
        if len(file_name.split(".")[0].split("_")[-1]) == 1:
            file_name = file_name.split("_")[0] + "_0" + file_name.split("_")[1]
            count_if_zeros_are_prefixed += 1
        files_list_ordered.append(file_name)

    files_list_ordered.sort()
    # the list is now in right order, but the 0 has to be deleted
    files_list_correct = []
    for file_name in files_list_ordered:
        if file_name.split(".")[0].split("_")[-1][0] == "0":
            file_name = file_name.split("_")[0] + "_" + file_name.split("_0")[-1]
        files_list_correct.append(file_name)

    if count_if_zeros_are_prefixed >= 5:
        # check if file names from marktstammdatenregister have no prefixed 0 already
        files_list = files_list_correct

    return files_list


def read_xml_file(f: ZipFile, file_name: str) -> pd.DataFrame:
    """Read the xml file from the zip file and return it as a DataFrame."""
    with f.open(file_name) as xml_file:
        try:
            return pd.read_xml(xml_file, encoding="UTF-16", parser="etree")
        except lxml.etree.XMLSyntaxError as error:
            return handle_xml_syntax_error(xml_file.read().decode("utf-16"), error)


def change_column_names_to_orm_format(
    df: pd.DataFrame, xml_table_name: str
) -> pd.DataFrame:
    if tablename_mapping[xml_table_name]["replace_column_names"]:
        df.rename(
            columns=tablename_mapping[xml_table_name]["replace_column_names"],
            inplace=True,
        )
    return df


def add_table_to_non_sqlite_database(
    df: pd.DataFrame,
    xml_table_name: str,
    sql_table_name: str,
    engine: sqlalchemy.engine.Engine,
) -> None:
    # get a dictionary for the data types
    table_columns_list = list(
        tablename_mapping[xml_table_name]["__class__"].__table__.columns
    )
    dtypes_for_writing_sql = {
        column.name: column.type
        for column in table_columns_list
        if column.name in df.columns
    }

    # Convert date and datetime columns into the datatype datetime.
    df = cast_date_columns_to_datetime(xml_table_name, df)

    add_missing_columns_to_table(
        engine, xml_table_name, column_list=df.columns.tolist()
    )

    for _ in range(10000):
        try:
            with engine.connect() as con:
                with con.begin():
                    df.to_sql(
                        sql_table_name,
                        con=con,
                        index=False,
                        if_exists="append",
                        dtype=dtypes_for_writing_sql,
                    )
                    break

        except sqlalchemy.exc.DataError as err:
            delete_wrong_xml_entry(err, df)

        except sqlalchemy.exc.IntegrityError:
            # error resulting from Unique constraint failed
            df = write_single_entries_until_not_unique_comes_up(
                df, xml_table_name, engine
            )


def add_zero_as_first_character_for_too_short_string(df: pd.DataFrame) -> pd.DataFrame:
    """Some columns are read as integer even though they are actually strings starting with
    a 0. This function converts those columns back to strings and adds a 0 as first character.
    """

    dict_of_columns_and_string_length = {
        "Gemeindeschluessel": 8,
        "Postleitzahl": 5,
    }
    for column_name, string_length in dict_of_columns_and_string_length.items():
        if column_name not in df.columns:
            continue
        try:
            df[column_name] = df[column_name].astype("Int64").astype(str)
        except (ValueError, TypeError):
            # some Plz are in the format DK-9999 for danish Postleitzahl
            # or A-9999 for austrian PLz
            # They cannot be converted to integer
            df[column_name] = df[column_name].astype(str)
            continue
        df[column_name] = df[column_name].where(
            cond=-df[column_name].isin(["None", "<NA>"]), other=None
        )

        string_adding_series = pd.Series(["0"] * len(df))
        string_adding_series = string_adding_series.where(
            cond=df[column_name].str.len() == string_length - 1, other=""
        )
        df[column_name] = string_adding_series + df[column_name]
    return df


def write_single_entries_until_not_unique_comes_up(
    df: pd.DataFrame, xml_table_name: str, engine: sqlalchemy.engine.Engine
) -> pd.DataFrame:
    """
    Remove from dataframe these rows, which are already existing in the database table
    Parameters
    ----------
    df
    xml_table_name
    engine

    Returns
    -------
    Filtered dataframe
    """

    table = tablename_mapping[xml_table_name]["__class__"].__table__
    primary_key = next(c for c in table.columns if c.primary_key)

    with engine.connect() as con:
        with con.begin():
            key_list = (
                pd.read_sql(sql=select(primary_key), con=con).values.squeeze().tolist()
            )

    len_df_before = len(df)
    df = df.drop_duplicates(
        subset=[primary_key.name]
    )  # drop all entries with duplicated primary keys in the dataframe
    df = df.set_index(primary_key.name)

    df = df.drop(
        labels=key_list, errors="ignore"
    )  # drop primary keys that already exist in the table
    df = df.reset_index()
    print(f"{len_df_before - len(df)} entries already existed in the database.")

    return df


def add_missing_columns_to_table(
    engine: sqlalchemy.engine.Engine,
    xml_table_name: str,
    column_list: list,
) -> None:
    """
    Some files introduce new columns for existing tables.
    If the pandas dataframe contains columns that do not
    exist in the database, they are added to the database.
    Parameters
    ----------
    engine
    xml_table_name
    column_list

    Returns
    -------

    """
    log = setup_logger()

    # get the columns name from the existing database
    inspector = sqlalchemy.inspect(engine)
    table_name = tablename_mapping[xml_table_name]["__class__"].__table__.name
    columns = inspector.get_columns(table_name)
    column_names_from_database = [column["name"] for column in columns]

    missing_columns = set(column_list) - set(column_names_from_database)

    for column_name in missing_columns:
        if not column_exists(engine, table_name, column_name):
            alter_query = 'ALTER TABLE %s ADD "%s" VARCHAR NULL;' % (
                table_name,
                column_name,
            )
            try:
                with engine.connect().execution_options(autocommit=True) as con:
                    with con.begin():
                        con.execute(
                            text(alter_query).execution_options(autocommit=True)
                        )
            except sqlalchemy.exc.OperationalError as err:
                # If the column already exists, we can ignore the error.
                if "duplicate column name" not in str(err):
                    raise err
            log.info(
                "From the downloaded xml files following new attribute was "
                f"introduced: {table_name}.{column_name}"
            )


def delete_wrong_xml_entry(err: Error, df: pd.DataFrame) -> pd.DataFrame:
    delete_entry = str(err).split("«")[0].split("»")[1]
    print(f"The entry {delete_entry} was deleted due to its false data type.")
    return df.replace(delete_entry, np.nan)


def handle_xml_syntax_error(data: str, err: Error) -> pd.DataFrame:
    """Deletes entries that cause an xml syntax error and produces DataFrame.

    Parameters
    -----------
    data : str
        Decoded xml file as one string
    err : ErrorMessage
        Error message that appeared when trying to use pd.read_xml on invalid xml file.

    Returns
    ----------
    df : pandas.DataFrame
        DataFrame which is read from the changed xml data.
    """

    def find_nearest_brackets(xml_string: str, position: int) -> tuple[int, int]:
        left_bracket_position = xml_string.rfind(">", 0, position)
        right_bracket_position = xml_string.find("<", position)
        return left_bracket_position, right_bracket_position

    data = data.splitlines()

    for _ in range(100):
        # check for maximum of 100 syntax errors, otherwise return an error
        wrong_char_row, wrong_char_column = err.position
        row_with_error = data[wrong_char_row - 1]

        left_bracket, right_bracket = find_nearest_brackets(
            row_with_error, wrong_char_column
        )
        data[wrong_char_row - 1] = (
            row_with_error[: left_bracket + 1] + row_with_error[right_bracket:]
        )
        try:
            print("One invalid xml expression was deleted.")
            df = pd.read_xml(StringIO("\n".join(data)))
            return df
        except lxml.etree.XMLSyntaxError as e:
            err = e
            continue

    raise Error("An error occured when parsing the xml file. Maybe it is corrupted?")


def process_table_before_insertion(
    df: pd.DataFrame,
    xml_table_name: str,
    zipped_xml_file_path: str,
    bulk_download_date: str,
    bulk_cleansing: bool,
) -> pd.DataFrame:
    df = add_zero_as_first_character_for_too_short_string(df)
    df = change_column_names_to_orm_format(df, xml_table_name)

    # Add Column that refers to the source of the data
    df["DatenQuelle"] = "bulk"
    df["DatumDownload"] = bulk_download_date

    if bulk_cleansing:
        df = cleanse_bulk_data(df, zipped_xml_file_path)
    return df


def add_table_to_sqlite_database(
    df: pd.DataFrame,
    xml_table_name: str,
    sql_table_name: str,
    engine: sqlalchemy.engine.Engine,
) -> None:
    column_list = df.columns.tolist()
    add_missing_columns_to_table(engine, xml_table_name, column_list)

    # Convert NaNs to None.
    df = df.where(pd.notnull(df), None)

    # Convert date columns to strings. Dates are not supported directly by SQLite.
    df = cast_date_columns_to_string(xml_table_name, df)

    # Create SQL statement for bulk insert. ON CONFLICT DO NOTHING prevents duplicates.
    insert_stmt = f"INSERT INTO {sql_table_name} ({','.join(column_list)}) VALUES ({','.join(['?' for _ in column_list])}) ON CONFLICT DO NOTHING"

    for _ in range(10000):
        try:
            with engine.connect() as con:
                with con.begin():
                    con.connection.executemany(insert_stmt, df.to_numpy())
                    break
        except sqlalchemy.exc.DataError as err:
            delete_wrong_xml_entry(err, df)
        except sqlalchemy.exc.IntegrityError:
            # error resulting from Unique constraint failed
            df = write_single_entries_until_not_unique_comes_up(
                df, xml_table_name, engine
            )
        except:
            # If any unexpected error occurs, we'll switch back to the non-SQLite method.
            add_table_to_non_sqlite_database(df, xml_table_name, sql_table_name, engine)
            break


def column_exists(engine, table_name, column_name):
    inspector = inspect(engine)
    columns = [col["name"] for col in inspector.get_columns(table_name)]
    return column_name in columns



import datetime
import os
import pandas as pd
from sqlalchemy import and_, func
from sqlalchemy.sql import exists
import shlex
import subprocess
from datetime import date

from open_mastr.utils.config import (
    setup_logger,
)
from open_mastr.soap_api.download import MaStRDownload, flatten_dict
from open_mastr.utils import orm
from open_mastr.utils.helpers import session_scope, reverse_unit_type_map

from open_mastr.utils.constants import ORM_MAP, UNIT_TYPE_MAP

log = setup_logger()


class MaStRMirror:
    """
    !!! warning

        **This class is deprecated** and will not be maintained in the future.
        Instead use [`Mastr.download`][open_mastr.Mastr.download] with parameter
        `method` = "bulk" to mirror the MaStR dataset to a local database.

    Mirror the Marktstammdatenregister database and keep it up-to-date.

    A PostgreSQL database is used to mirror the MaStR database. It builds
    on functionality for bulk data download
    provided by [`MaStRDownload`][open_mastr.soap_api.download.MaStRDownload].

    A rough overview is given by the following schema on the example of wind power units.
    ![Schema on the example of downloading wind power units using the API](../images/MaStR_Mirror.svg){ loading=lazy width=70%  align=center}



    Initially, basic unit data gets backfilled with `~.backfill_basic`
    (downloads basic unit data for 2,000
    units of type 'solar').

    ```python

       from open_mastr.soap_api.prototype_mastr_reflected import MaStRMirror

       mastr_mirror = MaStRMirror()
       mastr_mirror.backfill_basic("solar", limit=2000)
    ```
    Based on this, requests for
    additional data are created. This happens during backfilling basic data.
    But it is also possible to (re-)create
    requests for remaining additional data using
    [`MaStRMirror.create_additional_data_requests`][open_mastr.soap_api.mirror.MaStRMirror.create_additional_data_requests].

    ```python

       mastr_mirror.create_additional_data_requests("solar")
    ```

    Additional unit data, in the case of wind power this is extended data,
    EEG data and permit data, can be
    retrieved subsequently by `~.retrieve_additional_data`.

    ```python
        mastr_mirror.retrieve_additional_data("solar", ["unit_data"])
    ```

    The data can be joined to one table for each data type and exported to
    CSV files using [`Mastr.to_csv`][open_mastr.mastr.Mastr.to_csv].

    Also consider to use `~.dump` and `~.restore` for specific purposes.

    !!! Note
        This feature was built before the bulk download was offered at marktstammdatenregister.de.
        It can still be used to compare the two datasets received from the API and the bulk download.

    """

    def __init__(
        self,
        engine,
        restore_dump=None,
        parallel_processes=None,
    ):
        """
        Parameters
        ----------
        engine: sqlalchemy.engine.Engine
            database engine
        restore_dump: str or path-like, optional
            Save path of SQL dump file including filename.
            The database is restored from the SQL dump.
            Defaults to `None` which means nothing gets restored.
            Should be used in combination with `empty_schema=True`.
        parallel_processes: int
            Number of parallel processes used to download additional data.
            Defaults to `None`.
        """
        log.warning(
            """
            The `MaStRMirror` class is deprecated and will not be maintained in the future.
            To get a full table of the Marktstammdatenregister, use the open_mastr.Mastr.download
            method.

            If this change causes problems for you, please comment in this issue on github:
            https://github.com/OpenEnergyPlatform/open-MaStR/issues/487

            """
        )

        self._engine = engine

        # Associate downloader
        self.mastr_dl = MaStRDownload(parallel_processes=parallel_processes)

        # Restore database from a dump
        if restore_dump:
            self.restore(restore_dump)

        # Map technologies on ORMs
        self.orm_map = ORM_MAP

        # Map data and MaStR unit type
        # Map technologies on ORMs
        self.unit_type_map = UNIT_TYPE_MAP
        self.unit_type_map_reversed = reverse_unit_type_map()

    def backfill_basic(self, data=None, date=None, limit=10**8) -> None:
        """Backfill basic unit data.

        Fill database table 'basic_units' with data. It allows specification
        of which data should be retrieved via
        the described parameter options.

        Under the hood, [`MaStRDownload.basic_unit_data`][open_mastr.soap_api.download.MaStRDownload.basic_unit_data] is used.

        Parameters
        ----------
        data: list
            Specify data types for which data should be backfilled.

            * ['solar']: Backfill data for a single data type.
            * ['solar', 'wind'] (`list`):  Backfill data for multiple technologies given in a list.
        date: None, :class:`datetime.datetime`, str
            Specify backfill date from which on data is retrieved

            Only data with modification time stamp greater that `date` is retrieved.

            * `datetime.datetime(2020, 11, 27)`: Retrieve data which is newer
            than this time stamp
            * 'latest': Retrieve data which is newer than the newest data
            already in the table.
              It is aware of a different 'latest date' for each data.
              Hence, it works in combination with
              `data=None` and `data=["wind", "solar"]` for example.

              !!! warning

                 Don't use 'latest' in combination with `limit`. This might
                 lead to unexpected results.
            * `None`: Complete backfill
        limit: int
            Maximum number of units.
            Defaults to the large number of 10**8 which means
            all available data is queried. Use with care!
        """

        dates = self._get_list_of_dates(date, data)

        for data_type, date in zip(data, dates):
            self._write_basic_data_for_one_data_type_to_db(data_type, date, limit)

    def backfill_locations_basic(
        self, limit=10**7, date=None, delete_additional_data_requests=True
    ):
        """
        Backfill basic location data.

        Fill database table 'locations_basic' with data. It allows specification
        of which data should be retrieved via
        the described parameter options.

        Under the hood, [`MaStRDownload.basic_location_data`][open_mastr.soap_api.download.MaStRDownload.basic_location_data]
        is used.

        Parameters
        ----------
        date: None, `datetime.datetime`, str
            Specify backfill date from which on data is retrieved

            Only data with modification time stamp greater that `date` is retrieved.

            * `datetime.datetime(2020, 11, 27)`: Retrieve data which is newer than
            this time stamp
            * 'latest': Retrieve data which is newer than the newest data already in the table.
              !!! warning

                 Don't use 'latest' in combination with `limit`. This might lead to
                 unexpected results.
            * `None`: Complete backfill
        limit: int
            Maximum number of locations to download.
            Defaults to `None` which means no limit is set and all available data is queried.
            Use with care!
        delete_additional_data_requests: bool
            Useful to speed up download of data. Ignores existence of already created requests
            for additional data and
            skips deletion these.
        """

        date = self._get_date(date, technology_list=None)
        locations_basic = self.mastr_dl.basic_location_data(limit, date_from=date)

        for locations_chunk in locations_basic:
            # Remove duplicates returned from API
            locations_chunk_unique = [
                location
                for n, location in enumerate(locations_chunk)
                if location["LokationMastrNummer"]
                not in [_["LokationMastrNummer"] for _ in locations_chunk[n + 1 :]]
            ]
            locations_unique_ids = [
                _["LokationMastrNummer"] for _ in locations_chunk_unique
            ]

            with session_scope(engine=self._engine) as session:
                # Find units that are already in the DB
                common_ids = [
                    _.LokationMastrNummer
                    for _ in session.query(
                        orm.LocationBasic.LokationMastrNummer
                    ).filter(
                        orm.LocationBasic.LokationMastrNummer.in_(locations_unique_ids)
                    )
                ]
                inserted_and_updated = self._create_inserted_and_updated_list(
                    "locations", session, locations_chunk_unique, common_ids
                )

                # Create data requests for all newly inserted and updated locations
                new_requests = [
                    {
                        "LokationMastrNummer": location["LokationMastrNummer"],
                        "location_type": self.unit_type_map[location["Lokationtyp"]],
                        "request_date": datetime.datetime.now(tz=datetime.timezone.utc),
                    }
                    for location in inserted_and_updated
                ]

                # Delete old data requests
                if delete_additional_data_requests:
                    ids_to_delete = [
                        _["LokationMastrNummer"] for _ in inserted_and_updated
                    ]
                    session.query(orm.AdditionalLocationsRequested).filter(
                        orm.AdditionalLocationsRequested.LokationMastrNummer.in_(
                            ids_to_delete
                        )
                    ).filter(
                        orm.AdditionalLocationsRequested.request_date
                        < datetime.datetime.now(tz=datetime.timezone.utc)
                    ).delete(
                        synchronize_session="fetch"
                    )
                    session.commit()

                # Do bulk insert of new data requests
                session.bulk_insert_mappings(
                    orm.AdditionalLocationsRequested, new_requests
                )

    def retrieve_additional_data(self, data, data_type, limit=10**8, chunksize=1000):
        """
        Retrieve additional unit data

        Execute additional data requests stored in
        `open_mastr.soap_api.orm.AdditionalDataRequested`.
        See also docs of [`MaStRDownload.additional_data`][open_mastr.soap_api.download.MaStRDownload.additional_data]
        for more information on how data is downloaded.

        Parameters
        ----------
        data: `str`
            See list of available technologies in
            `open_mastr.soap_api.download.py.MaStRDownload.download_power_plants`.
        data_type: `str`
            Select type of additional data that is to be retrieved. Choose from
            "unit_data", "eeg_data", "kwk_data", "permit_data".
        limit: int
            Limit number of units that data is download for. Defaults to the very large number 10**8
            which refers to query data for existing data requests, for example created by
            `~.create_additional_data_requests`.
        chunksize: int
            Data is downloaded and inserted into the database in chunks of `chunksize`.
            Defaults to 1000.
        """

        # Mapping of download from MaStRDownload
        download_functions = {
            "unit_data": "extended_unit_data",
            "eeg_data": "eeg_unit_data",
            "kwk_data": "kwk_unit_data",
            "permit_data": "permit_unit_data",
        }

        if chunksize > limit:
            chunksize = limit

        number_units_queried = 0
        while number_units_queried < limit:
            with session_scope(engine=self._engine) as session:
                (
                    requested_chunk,
                    requested_ids,
                ) = self._get_additional_data_requests_from_db(
                    table_identifier="additional_data",
                    session=session,
                    data_request_type=data_type,
                    data=data,
                    chunksize=chunksize,
                )

                if not requested_ids:
                    log.info("No further data is requested")
                    break

                # Retrieve data
                unit_data, missed_units = self.mastr_dl.additional_data(
                    data, requested_ids, download_functions[data_type]
                )

                unit_data = flatten_dict(unit_data, serialize_with_json=False)
                number_units_merged = 0

                # Prepare data and add to database table
                for unit_dat in unit_data:
                    unit = self._preprocess_additional_data_entry(
                        unit_dat, data, data_type
                    )
                    session.merge(unit)
                    number_units_merged += 1
                session.commit()

                log.info(
                    f"Downloaded data for {len(unit_data)} units ({len(requested_ids)} requested). "
                )
                self._delete_missed_data_from_request_table(
                    table_identifier="additional_data",
                    session=session,
                    missed_requests=missed_units,
                    requested_chunk=requested_chunk,
                )
                # Update while iteration condition
                number_units_queried += len(requested_ids)
            # Emergency break out: if now new data gets inserted/update, don't retrieve any
            # further data
            if number_units_merged == 0:
                log.info("No further data is requested")
                break

    def retrieve_additional_location_data(
        self, location_type, limit=10**8, chunksize=1000
    ):
        """
        Retrieve extended location data

        Execute additional data requests stored in
        `open_mastr.soap_api.orm.AdditionalLocationsRequested`.
        See also docs of `open_mastr.soap_api.download.py.MaStRDownload.additional_data`
        for more information on how data is downloaded.

        Parameters
        ----------
        location_type: `str`
            Select type of location that is to be retrieved. Choose from
            "location_elec_generation", "location_elec_consumption", "location_gas_generation",
            "location_gas_consumption".
        limit: int
            Limit number of locations that data is download for. Defaults large number 10**8
            which refers to query data for existing data requests.
        chunksize: int
            Data is downloaded and inserted into the database in chunks of `chunksize`.
            Defaults to 1000.
        """

        # Process arguments
        if chunksize > limit:
            chunksize = limit

        locations_queried = 0
        while locations_queried < limit:
            with session_scope(engine=self._engine) as session:
                # Get a chunk
                (
                    requested_chunk,
                    requested_ids,
                ) = self._get_additional_data_requests_from_db(
                    table_identifier="additional_location_data",
                    session=session,
                    data_request_type=location_type,
                    data=None,
                    chunksize=chunksize,
                )

                if not requested_ids:
                    log.info("No further data is requested")
                    break

                # Reset number of locations inserted or updated for this chunk
                number_locations_merged = 0

                # Retrieve data
                location_data, missed_locations = self.mastr_dl.additional_data(
                    location_type, requested_ids, "location_data"
                )

                # Prepare data and add to database table
                location_data = flatten_dict(location_data)
                for location_dat in location_data:
                    location_dat = self._add_data_source_and_download_date(location_dat)
                    # Remove query status information from response
                    for exclude in [
                        "Ergebniscode",
                        "AufrufVeraltet",
                        "AufrufVersion",
                        "AufrufLebenszeitEnde",
                    ]:
                        del location_dat[exclude]

                    # Make data types JSON serializable
                    location_dat["DatumLetzteAktualisierung"] = location_dat[
                        "DatumLetzteAktualisierung"
                    ].isoformat()

                    if type(location_dat["DatumLetzteAktualisierung"]) == str:
                        location_dat[
                            "DatumLetzteAktualisierung"
                        ] = datetime.datetime.strptime(
                            location_dat["DatumLetzteAktualisierung"],
                            "%Y-%m-%dT%H:%M:%S.%f",
                        )

                    # Create new instance and update potentially existing one
                    location = orm.LocationExtended(**location_dat)
                    session.merge(location)
                    number_locations_merged += 1

                    session.commit()
                    # Log locations where data retrieval was not successful
                    log.info(f"Downloaded data for {len(location_data)} ")
                    self._delete_missed_data_from_request_table(
                        table_identifier="additional_location_data",
                        session=session,
                        missed_requests=missed_locations,
                        requested_chunk=requested_chunk,
                    )
                    # Update while iteration condition
                    locations_queried += len(requested_ids)

            # Emergency break out: if now new data gets inserted/update,
            # don't retrieve any further data
            if number_locations_merged == 0:
                log.info("No further data is requested")
                break

    def create_additional_data_requests(
        self,
        technology,
        data_types=["unit_data", "eeg_data", "kwk_data", "permit_data"],
        delete_existing=True,
    ):
        """
        Create new requests for additional unit data

        For units that exist in basic_units but not in the table for additional
        data of `data_type`, a new data request
        is submitted.

        Parameters
        ----------
        technology: str
            Specify technology additional data should be requested for.
        data_types: list
            Select type of additional data that is to be requested.
            Defaults to all data that is available for a
            technology.
        delete_existing: bool
            Toggle deletion of already existing requests for additional data.
            Defaults to True.
        """

        data_requests = []

        with session_scope(engine=self._engine) as session:
            # Check which additional data is missing
            for data_type in data_types:
                if data_type_available := self.orm_map[technology].get(data_type, None):
                    log.info(
                        f"Create requests for additional data of type {data_type} for {technology}"
                    )

                    # Get ORM for additional data by technology and data_type
                    additional_data_orm = getattr(orm, data_type_available)

                    # Delete prior additional data requests for this technology and data_type
                    if delete_existing:
                        session.query(orm.AdditionalDataRequested).filter(
                            orm.AdditionalDataRequested.technology == technology,
                            orm.AdditionalDataRequested.data_type == data_type,
                        ).delete()
                        session.commit()

                    # Query database for missing additional data
                    units_for_request = self._get_units_for_request(
                        data_type, session, additional_data_orm, technology
                    )

                    # Prepare data for additional data request
                    for basic_unit in units_for_request:
                        data_request = {
                            "EinheitMastrNummer": basic_unit.EinheitMastrNummer,
                            "technology": self.unit_type_map[basic_unit.Einheittyp],
                            "data_type": data_type,
                            "request_date": datetime.datetime.now(
                                tz=datetime.timezone.utc
                            ),
                        }
                        if data_type == "unit_data":
                            data_request[
                                "additional_data_id"
                            ] = basic_unit.EinheitMastrNummer
                        elif data_type == "eeg_data":
                            data_request[
                                "additional_data_id"
                            ] = basic_unit.EegMastrNummer
                        elif data_type == "kwk_data":
                            data_request[
                                "additional_data_id"
                            ] = basic_unit.KwkMastrNummer
                        elif data_type == "permit_data":
                            data_request[
                                "additional_data_id"
                            ] = basic_unit.GenMastrNummer
                        data_requests.append(data_request)

            # Insert new requests for additional data into database
            session.bulk_insert_mappings(orm.AdditionalDataRequested, data_requests)

    def _add_data_source_and_download_date(self, entry: dict) -> dict:
        """Adds DatenQuelle = 'APT' and DatumDownload = date.today"""
        entry["DatenQuelle"] = "API"
        entry["DatumDownload"] = date.today()
        return entry

    def _create_data_list_from_basic_units(self, session, basic_units_chunk):
        # Make sure that no duplicates get inserted into database
        # (would result in an error)
        # Only new data gets inserted or data with newer modification date gets updated

        # Remove duplicates returned from API
        basic_units_chunk_unique = [
            unit
            for n, unit in enumerate(basic_units_chunk)
            if unit["EinheitMastrNummer"]
            not in [_["EinheitMastrNummer"] for _ in basic_units_chunk[n + 1 :]]
        ]
        basic_units_chunk_unique_ids = [
            _["EinheitMastrNummer"] for _ in basic_units_chunk_unique
        ]

        # Find units that are already in the DB
        common_ids = [
            _.EinheitMastrNummer
            for _ in session.query(orm.BasicUnit.EinheitMastrNummer).filter(
                orm.BasicUnit.EinheitMastrNummer.in_(basic_units_chunk_unique_ids)
            )
        ]
        basic_units_chunk_unique = self._correct_typo_in_column_name(
            basic_units_chunk_unique
        )

        inserted_and_updated = self._create_inserted_and_updated_list(
            "basic_units", session, basic_units_chunk_unique, common_ids
        )

        # Submit additional data requests

        extended_data = []
        eeg_data = []
        kwk_data = []
        permit_data = []

        for basic_unit in inserted_and_updated:
            extended_data = self._append_additional_data_from_basic_unit(
                extended_data, basic_unit, "EinheitMastrNummer", "unit_data"
            )
            eeg_data = self._append_additional_data_from_basic_unit(
                eeg_data, basic_unit, "EegMastrNummer", "eeg_data"
            )
            kwk_data = self._append_additional_data_from_basic_unit(
                kwk_data, basic_unit, "KwkMastrNummer", "kwk_data"
            )
            permit_data = self._append_additional_data_from_basic_unit(
                permit_data, basic_unit, "GenMastrNummer", "permit_data"
            )
        return extended_data, eeg_data, kwk_data, permit_data, inserted_and_updated

    def _correct_typo_in_column_name(self, basic_units_chunk_unique: list) -> list:
        """
        Corrects the typo DatumLetzeAktualisierung -> DatumLetzteAktualisierung
        (missing t in Letzte) in the column name.
        """
        basic_units_chunk_unique_correct = []
        for unit in basic_units_chunk_unique:
            # Rename the typo in column DatumLetzeAktualisierung
            if "DatumLetzteAktualisierung" not in unit:
                unit["DatumLetzteAktualisierung"] = unit.pop(
                    "DatumLetzeAktualisierung", None
                )
            basic_units_chunk_unique_correct.append(unit)
        return basic_units_chunk_unique_correct

    def _append_additional_data_from_basic_unit(
        self,
        data_list: list,
        basic_unit: dict,
        basic_unit_identifier: str,
        data_type: str,
    ) -> list:
        """Appends a new entry from basic units to an existing list of unit IDs. This list is
        used when requesting additional data from the MaStR API."""
        if basic_unit[basic_unit_identifier]:
            data_list.append(
                {
                    "EinheitMastrNummer": basic_unit["EinheitMastrNummer"],
                    "additional_data_id": basic_unit[basic_unit_identifier],
                    "technology": self.unit_type_map[basic_unit["Einheittyp"]],
                    "data_type": data_type,
                    "request_date": datetime.datetime.now(tz=datetime.timezone.utc),
                }
            )
        return data_list

    def _create_inserted_and_updated_list(
        self, table_identifier, session, list_chunk_unique, common_ids
    ) -> list:
        """Creates the insert and update list and saves it to the BasicTable.
        This method is called both in backfill_basics and backfill_location_basics."""
        if table_identifier == "locations":
            mastr_number_identifier = "LokationMastrNummer"
            table_class = orm.LocationBasic
        elif table_identifier == "basic_units":
            mastr_number_identifier = "EinheitMastrNummer"
            table_class = orm.BasicUnit

        insert = []
        updated = []
        for entry in list_chunk_unique:
            # In case data for the unit already exists, only update if new data is newer
            if entry[mastr_number_identifier] in common_ids:
                query_filter = (
                    and_(
                        orm.BasicUnit.EinheitMastrNummer == entry["EinheitMastrNummer"],
                        orm.BasicUnit.DatumLetzteAktualisierung
                        < entry["DatumLetzteAktualisierung"],
                    )
                    if table_identifier == "basic_units"
                    else orm.LocationBasic.LokationMastrNummer
                    == entry["LokationMastrNummer"]
                )
                if session.query(exists().where(query_filter)).scalar():
                    updated.append(entry)
                    session.merge(table_class(**entry))
            # In case of new data, just insert
            else:
                insert.append(entry)
        session.bulk_save_objects([table_class(**u) for u in insert])
        session.commit()
        return insert + updated

    def _write_basic_data_for_one_data_type_to_db(self, data, date, limit) -> None:
        log.info(f"Backfill data for data type {data}")

        # Catch weird MaStR SOAP response
        basic_units = self.mastr_dl.basic_unit_data(data, limit, date_from=date)

        with session_scope(engine=self._engine) as session:
            log.info(
                "Insert basic unit data into DB and submit additional data requests"
            )
            for basic_units_chunk in basic_units:
                # Insert basic data into database
                (
                    extended_data,
                    eeg_data,
                    kwk_data,
                    permit_data,
                    inserted_and_updated,
                ) = self._create_data_list_from_basic_units(session, basic_units_chunk)

                # Delete old entries for additional data requests
                additional_data_table = orm.AdditionalDataRequested.__table__
                ids_to_delete = [_["EinheitMastrNummer"] for _ in inserted_and_updated]
                session.execute(
                    additional_data_table.delete()
                    .where(
                        additional_data_table.c.EinheitMastrNummer.in_(ids_to_delete)
                    )
                    .where(additional_data_table.c.technology == "wind")
                    .where(
                        additional_data_table.c.request_date
                        < datetime.datetime.now(tz=datetime.timezone.utc)
                    )
                )

                # Flush delete statements to database
                session.commit()

                # Insert new requests for additional data
                session.bulk_insert_mappings(orm.AdditionalDataRequested, extended_data)
                session.bulk_insert_mappings(orm.AdditionalDataRequested, eeg_data)
                session.bulk_insert_mappings(orm.AdditionalDataRequested, kwk_data)
                session.bulk_insert_mappings(orm.AdditionalDataRequested, permit_data)

            log.info("Backfill successfully finished")

    def _get_date(self, date, technology_list):
        """Parses 'latest' to the latest date in the database, else returns the given date."""
        if technology_list:
            return self._get_list_of_dates(date, technology_list)
        else:
            return self._get_single_date(date)

    def _get_single_date(self, date):
        if date != "latest":
            return date

        with session_scope(engine=self._engine) as session:
            if date_queried := (
                session.query(orm.LocationExtended.DatumLetzteAktualisierung)
                .order_by(orm.LocationExtended.DatumLetzteAktualisierung.desc())
                .first()
            ):
                return date_queried[0]
            else:
                return None

    def _get_list_of_dates(self, date, technology_list) -> list:
        if date != "latest":
            return [date] * len(technology_list)

        dates = []
        for tech in technology_list:
            if tech:
                # In case technologies are specified, latest data date
                # gets queried per data
                with session_scope(engine=self._engine) as session:
                    newest_date = (
                        session.query(orm.BasicUnit.DatumLetzteAktualisierung)
                        .filter(
                            orm.BasicUnit.Einheittyp
                            == self.unit_type_map_reversed[tech]
                        )
                        .order_by(orm.BasicUnit.DatumLetzteAktualisierung.desc())
                        .first()
                    )
            else:
                # If technologies aren't defined ([None]) latest date per data
                #  is queried in query
                # This also leads that the remainder of the loop body is skipped
                with session_scope(engine=self._engine) as session:
                    subquery = session.query(
                        orm.BasicUnit.Einheittyp,
                        func.max(orm.BasicUnit.DatumLetzteAktualisierung).label(
                            "maxdate"
                        ),
                    ).group_by(orm.BasicUnit.Einheittyp)
                    dates = [s[1] for s in subquery]
                    technology_list = [self.unit_type_map[s[0]] for s in subquery]
                    # Break the for loop over data here, because we
                    # write technology_list and dates at once
                    break

            # Add date to dates list
            if newest_date:
                dates.append(newest_date[0])
            # Cover the case where no data is in the database and latest is still used
            else:
                dates.append(None)

        return dates

    def _delete_missed_data_from_request_table(
        self, table_identifier, session, missed_requests, requested_chunk
    ):
        if table_identifier == "additional_data":
            id_attribute = "additional_data_id"
        elif table_identifier == "additional_location_data":
            id_attribute = "LokationMastrNummer"

        missed_entry_ids = [e[0] for e in missed_requests]
        for missed_req in missed_requests:
            missed = (
                orm.MissedAdditionalData(
                    additional_data_id=missed_req[0], reason=missed_req[1]
                )
                if table_identifier == "additional_data"
                else orm.MissedExtendedLocation(
                    LokationMastrNummer=missed_req[0],
                    reason=missed_req[1],
                )
            )
            session.add(missed)
        session.commit()
        # Remove entries from additional data request table if additional data
        # was retrieved
        deleted_entries = []
        for requested_entry in requested_chunk:
            if getattr(requested_entry, id_attribute) not in missed_entry_ids:
                session.delete(requested_entry)
                deleted_entries.append(getattr(requested_entry, id_attribute))
        log.info(
            f"Missed requests: {len(missed_requests)}. "
            f"Deleted requests: {len(deleted_entries)}."
        )
        session.commit()

    def _preprocess_additional_data_entry(self, unit_dat, technology, data_type):
        unit_dat = self._add_data_source_and_download_date(unit_dat)
        # Remove query status information from response
        for exclude in [
            "Ergebniscode",
            "AufrufVeraltet",
            "AufrufVersion",
            "AufrufLebenszeitEnde",
        ]:
            del unit_dat[exclude]

        # Pre-serialize dates/datetimes and decimal in hydro Ertuechtigung
        # This is required because sqlalchemy does not know how serialize
        # dates/decimal of a JSON
        if "Ertuechtigung" in unit_dat:
            for ertuechtigung in unit_dat["Ertuechtigung"]:
                if ertuechtigung["DatumWiederinbetriebnahme"]:
                    ertuechtigung["DatumWiederinbetriebnahme"] = ertuechtigung[
                        "DatumWiederinbetriebnahme"
                    ].isoformat()
                ertuechtigung["ProzentualeErhoehungDesLv"] = float(
                    ertuechtigung["ProzentualeErhoehungDesLv"]
                )
        # Some data (data_in_list) is handed over as type:list, hence
        # non-compatible with sqlite or postgresql
        # This replaces the list with the first element in the list

        data_as_list = ["NetzbetreiberMastrNummer", "Netzbetreiberzuordnungen"]

        for dat in data_as_list:
            if dat in unit_dat and type(unit_dat[dat]) == list:
                if len(unit_dat[dat]) > 0:
                    unit_dat[dat] = f"{unit_dat[dat][0]}"
                else:
                    unit_dat[dat] = None

        # Rename the typo in column zugeordneteWirkleistungWechselrichter
        if "zugeordneteWirkleistungWechselrichter" in unit_dat.keys():
            unit_dat["ZugeordneteWirkleistungWechselrichter"] = unit_dat.pop(
                "zugeordneteWirkleistungWechselrichter"
            )

        # Create new instance and update potentially existing one
        return getattr(orm, self.orm_map[technology][data_type])(**unit_dat)

    def _get_additional_data_requests_from_db(
        self, table_identifier, session, data_request_type, data, chunksize
    ):
        """Retrieves the data that is requested from the database table AdditionalDataRequested."""
        if table_identifier == "additional_data":
            requested_chunk = (
                session.query(orm.AdditionalDataRequested)
                .filter(
                    and_(
                        orm.AdditionalDataRequested.data_type == data_request_type,
                        orm.AdditionalDataRequested.technology == data,
                    )
                )
                .limit(chunksize)
            )

            ids = [_.additional_data_id for _ in requested_chunk]
        if table_identifier == "additional_location_data":
            requested_chunk = (
                session.query(orm.AdditionalLocationsRequested)
                .filter(
                    orm.AdditionalLocationsRequested.location_type == data_request_type
                )
                .limit(chunksize)
            )
            ids = [_.LokationMastrNummer for _ in requested_chunk]
        return requested_chunk, ids

    def _get_units_for_request(
        self, data_type, session, additional_data_orm, technology
    ):
        if data_type == "unit_data":
            units_for_request = (
                session.query(orm.BasicUnit)
                .outerjoin(
                    additional_data_orm,
                    orm.BasicUnit.EinheitMastrNummer
                    == additional_data_orm.EinheitMastrNummer,
                )
                .filter(
                    orm.BasicUnit.Einheittyp == self.unit_type_map_reversed[technology]
                )
                .filter(additional_data_orm.EinheitMastrNummer.is_(None))
                .filter(orm.BasicUnit.EinheitMastrNummer.isnot(None))
            )
        elif data_type == "eeg_data":
            units_for_request = (
                session.query(orm.BasicUnit)
                .outerjoin(
                    additional_data_orm,
                    orm.BasicUnit.EegMastrNummer == additional_data_orm.EegMastrNummer,
                )
                .filter(
                    orm.BasicUnit.Einheittyp == self.unit_type_map_reversed[technology]
                )
                .filter(additional_data_orm.EegMastrNummer.is_(None))
                .filter(orm.BasicUnit.EegMastrNummer.isnot(None))
            )
        elif data_type == "kwk_data":
            units_for_request = (
                session.query(orm.BasicUnit)
                .outerjoin(
                    additional_data_orm,
                    orm.BasicUnit.KwkMastrNummer == additional_data_orm.KwkMastrNummer,
                )
                .filter(
                    orm.BasicUnit.Einheittyp == self.unit_type_map_reversed[technology]
                )
                .filter(additional_data_orm.KwkMastrNummer.is_(None))
                .filter(orm.BasicUnit.KwkMastrNummer.isnot(None))
            )
        elif data_type == "permit_data":
            units_for_request = (
                session.query(orm.BasicUnit)
                .outerjoin(
                    additional_data_orm,
                    orm.BasicUnit.GenMastrNummer == additional_data_orm.GenMastrNummer,
                )
                .filter(
                    orm.BasicUnit.Einheittyp == self.unit_type_map_reversed[technology]
                )
                .filter(additional_data_orm.GenMastrNummer.is_(None))
                .filter(orm.BasicUnit.GenMastrNummer.isnot(None))
            )
        else:
            raise ValueError(f"Data type {data_type} is not a valid option.")

        return units_for_request

    def dump(self, dumpfile="open-mastr-continuous-update.backup"):
        """
        Dump MaStR database.

        Parameters
        ----------
        dumpfile : str or path-like, optional
            Save path for dump including filename. When only a filename is given,
            the dump is saved to CWD.
        """
        dump_cmd = (
            f"pg_dump -Fc "
            f"-f {dumpfile} "
            f"-n mastr_mirrored "
            f"-h localhost "
            f"-U open-mastr "
            f"-p 55443 "
            f"open-mastr"
        )

        proc = subprocess.Popen(dump_cmd, shell=True, env={"PGPASSWORD": "open-mastr"})
        proc.wait()

    def restore(self, dumpfile):
        """
        Restore the MaStR database from an SQL dump.

        Parameters
        ----------
        dumpfile : str or path-like, optional
            Save path for dump including filename. When only a filename is given, the
            dump is restored from CWD.


        !!! warning

            If tables that are restored from the dump contain data, restore doesn't work!

        """
        # Interpret file name and path
        dump_file_dir, dump_file = os.path.split(dumpfile)
        cwd = os.path.abspath(os.path.dirname(dump_file_dir))

        # Define import of SQL dump with pg_restore
        restore_cmd = (
            f"pg_restore -h localhost -U open-mastr -p 55443 -d open-mastr {dump_file}"
        )
        restore_cmd = shlex.split(restore_cmd)

        # Execute restore command
        proc = subprocess.Popen(
            restore_cmd,
            shell=False,
            env={"PGPASSWORD": "open-mastr"},
            cwd=cwd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        proc.wait()


def list_of_dicts_to_columns(row) -> pd.Series:  # FIXME: Function not used
    """
    Expand data stored in dict to spearate columns

    Parameters
    ----------
    row: list of dict
        Usually applied using apply on a column of a pandas DataFrame,
        hence, a Series. This column of the
        DataFrame should consist of a single-level dict with an
        arbitrary number of columns. Each key is
        transformed into a new column, while data from
        each dict inside the list is concatenated by key. Such
        that the data is stored into a list for each key/column.

    Returns
    -------
    pd.Series
        Pandas Series with keys as columns and values concatenated to a list for each key.
    """
    columns = {k: [] for dic in row for k, _ in dic.items()}
    for dic in row:
        for k, v in dic.items():
            columns[k].append(v)

    return pd.Series(columns)



import pandas as pd
import numpy as np
from open_mastr.xml_download.colums_to_replace import (
    system_catalog,
    columns_replace_list,
)
from zipfile import ZipFile


def cleanse_bulk_data(df: pd.DataFrame, zipped_xml_file_path: str) -> pd.DataFrame:
    df = replace_ids_with_names(df, system_catalog)
    # Katalogeintraege: int -> string value
    df = replace_mastr_katalogeintraege(
        zipped_xml_file_path=zipped_xml_file_path, df=df
    )
    return df


def replace_ids_with_names(df: pd.DataFrame, system_catalog: dict) -> pd.DataFrame:
    """Replaces ids with names according to the system catalog. This is
    necessary since the data from the bulk download encodes columns with
    IDs instead of the actual values."""
    for column_name, name_mapping_dictionary in system_catalog.items():
        if column_name in df.columns:
            df[column_name] = df[column_name].replace(name_mapping_dictionary)
    return df


def replace_mastr_katalogeintraege(
    zipped_xml_file_path: str,
    df: pd.DataFrame,
) -> pd.DataFrame:
    """Replaces the IDs from the mastr database by its mapped string values from
    the table katalogwerte"""
    katalogwerte = create_katalogwerte_from_bulk_download(zipped_xml_file_path)
    for column_name in df.columns:
        if column_name in columns_replace_list:
            if df[column_name].dtype == "O":
                # Handle comma seperated strings from catalog values
                df[column_name] = (
                    df[column_name]
                    .str.split(",", expand=True)
                    .apply(lambda x: x.str.strip())
                    .replace("", None)
                    .astype("Int64")
                    .map(katalogwerte.get)
                    .agg(lambda d: ",".join(i for i in d if isinstance(i, str)), axis=1)
                    .replace("", None)
                )
            else:
                df[column_name] = (
                    df[column_name].astype("float").astype("Int64").map(katalogwerte)
                )

    return df


def create_katalogwerte_from_bulk_download(zipped_xml_file_path) -> dict:
    """Creates a dictionary from the id -> value mapping defined in the table
    katalogwerte from MaStR."""
    with ZipFile(zipped_xml_file_path, "r") as f:
        data = f.read("Katalogwerte.xml")
        df_katalogwerte = pd.read_xml(data, encoding="UTF-16", compression="zip")
    katalogwerte_array = np.array(df_katalogwerte[["Id", "Wert"]])
    katalogwerte = dict(
        (katalogwerte_array[n][0], katalogwerte_array[n][1])
        for n in range(len(katalogwerte_array))
    )
    return katalogwerte



from open_mastr.utils.helpers import reverse_fill_basic_units, create_db_query


technology = [
    "solar",
    "wind",
    "biomass",
    "combustion",
    "gsgk",
    "hydro",
    "nuclear",
    "storage",
]

data_types = ["unit_data", "eeg_data", "kwk_data", "permit_data"]
location_types = [
    "location_elec_generation",
    "location_elec_consumption",
    "location_gas_generation",
    "location_gas_consumption",
]

# Fill the basic_units table from extended tables
reverse_fill_basic_units()

# to csv per tech
create_db_query(technology=technology, additional_data=data_types, limit=None)
